import re

import numpy as np
import pandas as pd
import pytest

from woodwork.datacolumn import DataColumn
from woodwork.exceptions import ColumnNameMismatchWarning, DuplicateTagsWarning
from woodwork.logical_types import (
    Categorical,
    CountryCode,
    Datetime,
    Double,
    Integer,
    NaturalLanguage,
    Ordinal,
    SubRegionCode,
    ZIPCode
)
from woodwork.tests.testing_utils import to_pandas
from woodwork.utils import import_or_none

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


def test_datacolumn_init(sample_series):
    data_col = DataColumn(sample_series, use_standard_tags=False)
    # Koalas doesn't support category dtype
    if not (ks and isinstance(sample_series, ks.Series)):
        sample_series = sample_series.astype('category')
    pd.testing.assert_series_equal(to_pandas(data_col.to_series()), to_pandas(sample_series))
    assert data_col.name == sample_series.name
    assert data_col.logical_type == Categorical
    assert data_col.semantic_tags == set()


def test_datacolumn_init_with_logical_type(sample_series):
    data_col = DataColumn(sample_series, NaturalLanguage)
    assert data_col.logical_type == NaturalLanguage
    assert data_col.semantic_tags == set()

    data_col = DataColumn(sample_series, "natural_language")
    assert data_col.logical_type == NaturalLanguage
    assert data_col.semantic_tags == set()

    data_col = DataColumn(sample_series, "NaturalLanguage")
    assert data_col.logical_type == NaturalLanguage
    assert data_col.semantic_tags == set()


def test_datacolumn_init_with_semantic_tags(sample_series):
    semantic_tags = ['tag1', 'tag2']
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)
    assert data_col.semantic_tags == set(semantic_tags)


def test_datacolumn_init_wrong_series():
    error = 'Series must be one of: pandas.Series, dask.Series, koalas.Series, or pandas.ExtensionArray'
    with pytest.raises(TypeError, match=error):
        DataColumn([1, 2, 3, 4])

    with pytest.raises(TypeError, match=error):
        DataColumn(np.array([1, 2, 3, 4]))


def test_datacolumn_init_with_name(sample_series, sample_datetime_series):
    name = 'sample_series'
    changed_name = 'changed_name'

    dc_use_series_name = DataColumn(sample_series)
    assert dc_use_series_name.name == name
    assert dc_use_series_name.to_series().name == name

    warning = 'Name mismatch between sample_series and changed_name. DataColumn and underlying series name are now changed_name'
    with pytest.warns(ColumnNameMismatchWarning, match=warning):
        dc_use_input_name = DataColumn(sample_series, name=changed_name)
    assert dc_use_input_name.name == changed_name
    assert dc_use_input_name.to_series().name == changed_name

    warning = 'Name mismatch between sample_datetime_series and changed_name. DataColumn and underlying series name are now changed_name'
    with pytest.warns(ColumnNameMismatchWarning, match=warning):
        dc_with_ltype_change = DataColumn(sample_datetime_series, name=changed_name)
    assert dc_with_ltype_change.name == changed_name
    assert dc_with_ltype_change.to_series().name == changed_name


def test_datacolumn_init_with_extension_array():
    series_categories = pd.Series([1, 2, 3], dtype='category')
    extension_categories = pd.Categorical([1, 2, 3])

    data_col = DataColumn(extension_categories)
    series = data_col.to_series()
    assert series.equals(series_categories)
    assert series.name is None
    assert data_col.name is None

    series_ints = pd.Series([1, 2, None, 4], dtype='Int64')
    extension_ints = pd.arrays.IntegerArray(np.array([1, 2, 3, 4], dtype="int64"), mask=np.array([False, False, True, False]))

    data_col_with_name = DataColumn(extension_ints, name='extension')
    series = data_col_with_name.to_series()
    assert series.equals(series_ints)
    assert series.name == 'extension'
    assert data_col_with_name.name == 'extension'

    series_strs = pd.Series([1, 2, None, 4], dtype='string')

    data_col_different_ltype = DataColumn(extension_ints, logical_type='NaturalLanguage')
    series = data_col_different_ltype.to_series()
    assert series.equals(series_strs)


def test_datacolumn_with_alternate_semantic_tags_input(sample_series):
    semantic_tags = 'custom_tag'
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)
    assert data_col.semantic_tags == {'custom_tag'}

    semantic_tags = {'custom_tag', 'numeric'}
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)
    assert data_col.semantic_tags == semantic_tags


def test_invalid_logical_type(sample_series):
    error_message = "Invalid logical type specified for 'sample_series'"
    with pytest.raises(TypeError, match=error_message):
        DataColumn(sample_series, int)

    error_message = "String naturalllanguage is not a valid logical type"
    with pytest.raises(ValueError, match=error_message):
        DataColumn(sample_series, 'naturalllanguage')


def test_semantic_tag_errors(sample_series):
    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        DataColumn(sample_series, semantic_tags=int)

    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        DataColumn(sample_series, semantic_tags={'index': {}, 'time_index': {}})

    error_message = "semantic_tags must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        DataColumn(sample_series, semantic_tags=['index', 1])


def test_datacolumn_description(sample_series):
    column_description = "custom description"
    data_col = DataColumn(sample_series, description=column_description)
    assert data_col.description == column_description

    new_description = "updated description text"
    data_col.description = new_description
    assert data_col.description == new_description


def test_datacolumn_description_error(sample_series):
    err_msg = "Column description must be a string"
    with pytest.raises(TypeError, match=err_msg):
        DataColumn(sample_series, description=123)


def test_datacolumn_repr(sample_series):
    data_col = DataColumn(sample_series, use_standard_tags=False)
    # Koalas doesn't support categorical
    if ks and isinstance(sample_series, ks.Series):
        dtype = 'object'
    else:
        dtype = 'category'
    assert data_col.__repr__() == f'<DataColumn: sample_series (Physical Type = {dtype}) ' \
        '(Logical Type = Categorical) (Semantic Tags = set())>'


def test_set_semantic_tags(sample_series):
    semantic_tags = {'tag1', 'tag2'}
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)
    assert data_col.semantic_tags == semantic_tags

    new_tags = ['new_tag']
    new_col = data_col.set_semantic_tags(new_tags)
    assert new_col is not data_col
    assert new_col.semantic_tags == set(new_tags)


def test_set_semantic_tags_with_index(sample_series):
    semantic_tags = {'tag1', 'tag2'}
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)
    data_col._set_as_index()
    assert data_col.semantic_tags == {'tag1', 'tag2', 'index'}
    new_tags = ['new_tag']
    new_col = data_col.set_semantic_tags(new_tags)
    assert new_col.semantic_tags == {'index', 'new_tag'}
    new_col2 = new_col.set_semantic_tags(new_tags, retain_index_tags=False)
    assert new_col2.semantic_tags == {'new_tag'}


def test_set_semantic_tags_with_time_index(sample_datetime_series):
    semantic_tags = {'tag1', 'tag2'}
    data_col = DataColumn(sample_datetime_series, semantic_tags=semantic_tags, use_standard_tags=False)
    data_col._set_as_time_index()
    assert data_col.semantic_tags == {'tag1', 'tag2', 'time_index'}
    new_tags = ['new_tag']
    new_col = data_col.set_semantic_tags(new_tags)
    assert new_col.semantic_tags == {'time_index', 'new_tag'}
    new_col2 = new_col.set_semantic_tags(new_tags, retain_index_tags=False)
    assert new_col2.semantic_tags == {'new_tag'}


def test_adds_numeric_standard_tag():
    series = pd.Series([1, 2, 3])
    semantic_tags = 'custom_tag'

    logical_types = [Integer, Double]
    for logical_type in logical_types:
        data_col = DataColumn(series, logical_type=logical_type, semantic_tags=semantic_tags)
        assert data_col.semantic_tags == {'custom_tag', 'numeric'}


def test_adds_category_standard_tag():
    series = pd.Series([1, 2, 3])
    semantic_tags = 'custom_tag'

    logical_types = [Categorical, CountryCode, Ordinal(order=(1, 2, 3)), SubRegionCode, ZIPCode]
    for logical_type in logical_types:
        data_col = DataColumn(series, logical_type=logical_type, semantic_tags=semantic_tags)
        assert data_col.semantic_tags == {'custom_tag', 'category'}


def test_does_not_add_standard_tags():
    series = pd.Series([1, 2, 3])
    semantic_tags = 'custom_tag'
    data_col = DataColumn(series,
                          logical_type=Double,
                          semantic_tags=semantic_tags,
                          use_standard_tags=False)
    assert data_col.semantic_tags == {'custom_tag'}


def test_add_custom_tags(sample_series):
    semantic_tags = 'initial_tag'
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)

    new_col = data_col.add_semantic_tags('string_tag')
    assert new_col is not data_col
    assert new_col.semantic_tags == {'initial_tag', 'string_tag'}

    new_col2 = new_col.add_semantic_tags(['list_tag'])
    assert new_col2.semantic_tags == {'initial_tag', 'string_tag', 'list_tag'}

    new_col3 = new_col2.add_semantic_tags({'set_tag'})
    assert new_col3.semantic_tags == {'initial_tag', 'string_tag', 'list_tag', 'set_tag'}


def test_warns_on_setting_duplicate_tag(sample_series):
    semantic_tags = ['first_tag', 'second_tag']
    data_col = DataColumn(sample_series, semantic_tags=semantic_tags, use_standard_tags=False)

    expected_message = "Semantic tag(s) 'first_tag, second_tag' already present on column 'sample_series'"
    with pytest.warns(DuplicateTagsWarning) as record:
        data_col.add_semantic_tags(['first_tag', 'second_tag'])
    assert len(record) == 1
    assert record[0].message.args[0] == expected_message


def test_set_logical_type_with_standard_tags(sample_series):
    data_col = DataColumn(sample_series,
                          logical_type=NaturalLanguage,
                          semantic_tags='original_tag',
                          use_standard_tags=True)

    new_col = data_col.set_logical_type(Categorical)
    assert isinstance(new_col, DataColumn)
    assert new_col is not data_col
    assert new_col.logical_type == Categorical
    assert new_col.semantic_tags == {'category'}


def test_set_logical_type_without_standard_tags(sample_series):
    data_col = DataColumn(sample_series,
                          logical_type=NaturalLanguage,
                          semantic_tags='original_tag',
                          use_standard_tags=False)

    new_col = data_col.set_logical_type(Categorical)
    assert isinstance(new_col, DataColumn)
    assert new_col is not data_col
    assert new_col.logical_type == Categorical
    assert new_col.semantic_tags == set()


def test_set_logical_type_retains_index_tag(sample_series):
    data_col = DataColumn(sample_series,
                          logical_type=NaturalLanguage,
                          semantic_tags='original_tag',
                          use_standard_tags=False)

    data_col._set_as_index()
    assert data_col.semantic_tags == {'index', 'original_tag'}
    new_col = data_col.set_logical_type(Categorical)
    assert new_col.semantic_tags == {'index'}
    new_col = data_col.set_logical_type(Categorical, retain_index_tags=False)
    assert new_col.semantic_tags == set()


def test_set_logical_type_retains_time_index_tag(sample_datetime_series):
    data_col = DataColumn(sample_datetime_series,
                          logical_type=Datetime,
                          semantic_tags='original_tag',
                          use_standard_tags=False)

    data_col._set_as_time_index()
    assert data_col.semantic_tags == {'time_index', 'original_tag'}
    new_col = data_col.set_logical_type(Categorical)
    assert new_col.semantic_tags == {'time_index'}
    new_col = data_col.set_logical_type(Categorical, retain_index_tags=False)
    assert new_col.semantic_tags == set()


def test_reset_semantic_tags_with_standard_tags(sample_series):
    semantic_tags = 'initial_tag'
    data_col = DataColumn(sample_series,
                          semantic_tags=semantic_tags,
                          logical_type=Categorical,
                          use_standard_tags=True)

    new_col = data_col.reset_semantic_tags()
    assert new_col is not data_col
    assert new_col.semantic_tags == Categorical.standard_tags


def test_reset_semantic_tags_without_standard_tags(sample_series):
    semantic_tags = 'initial_tag'
    data_col = DataColumn(sample_series,
                          semantic_tags=semantic_tags,
                          use_standard_tags=False)

    new_col = data_col.reset_semantic_tags()
    assert new_col is not data_col
    assert new_col.semantic_tags == set()


def test_reset_semantic_tags_with_index(sample_series):
    semantic_tags = 'initial_tag'
    data_col = DataColumn(sample_series,
                          semantic_tags=semantic_tags,
                          use_standard_tags=False)

    data_col._set_as_index()
    new_col = data_col.reset_semantic_tags(retain_index_tags=True)
    assert new_col.semantic_tags == {'index'}
    new_col = data_col.reset_semantic_tags()
    assert new_col.semantic_tags == set()


def test_reset_semantic_tags_with_time_index(sample_datetime_series):
    semantic_tags = 'initial_tag'
    data_col = DataColumn(sample_datetime_series,
                          semantic_tags=semantic_tags,
                          use_standard_tags=False)

    data_col._set_as_time_index()
    new_col = data_col.reset_semantic_tags(retain_index_tags=True)
    assert new_col.semantic_tags == {'time_index'}
    new_col = data_col.reset_semantic_tags()
    assert new_col.semantic_tags == set()


def test_remove_semantic_tags(sample_series):
    tags_to_remove = [
        'tag1',
        ['tag1'],
        {'tag1'}
    ]

    data_col = DataColumn(sample_series,
                          semantic_tags=['tag1', 'tag2'],
                          use_standard_tags=False)

    for tag in tags_to_remove:
        new_col = data_col.remove_semantic_tags(tag)
        assert new_col is not data_col
        assert new_col.semantic_tags == {'tag2'}


def test_remove_standard_semantic_tag(sample_series):
    # Check that warning is raised if use_standard_tags is True - tag should be removed
    data_col = DataColumn(sample_series,
                          logical_type=Categorical,
                          semantic_tags='tag1',
                          use_standard_tags=True)
    expected_message = "Removing standard semantic tag(s) 'category' from column 'sample_series'"
    with pytest.warns(UserWarning) as record:
        new_col = data_col.remove_semantic_tags(['tag1', 'category'])
    assert len(record) == 1
    assert record[0].message.args[0] == expected_message
    assert new_col.semantic_tags == set()

    # Check that warning is not raised if use_standard_tags is False - tag should be removed
    data_col = DataColumn(sample_series,
                          logical_type=Categorical,
                          semantic_tags=['category', 'tag1'],
                          use_standard_tags=False)

    with pytest.warns(None) as record:
        new_col = data_col.remove_semantic_tags(['tag1', 'category'])
    assert len(record) == 0
    assert new_col.semantic_tags == set()


def test_remove_semantic_tags_raises_error_with_invalid_tag(sample_series):
    data_col = DataColumn(sample_series,
                          semantic_tags='tag1')
    error_msg = re.escape("Semantic tag(s) 'invalid_tagname' not present on column 'sample_series'")
    with pytest.raises(LookupError, match=error_msg):
        data_col.remove_semantic_tags('invalid_tagname')


def test_raises_error_setting_index_tag_directly(sample_series):
    error_msg = re.escape("Cannot add 'index' tag directly. To set a column as the index, "
                          "use DataTable.set_index() instead.")
    with pytest.raises(ValueError, match=error_msg):
        DataColumn(sample_series, semantic_tags='index')

    data_col = DataColumn(sample_series)
    with pytest.raises(ValueError, match=error_msg):
        data_col.add_semantic_tags('index')
    with pytest.raises(ValueError, match=error_msg):
        data_col.set_semantic_tags('index')


def test_raises_error_setting_time_index_tag_directly(sample_series):
    error_msg = re.escape("Cannot add 'time_index' tag directly. To set a column as the time index, "
                          "use DataTable.set_time_index() instead.")
    with pytest.raises(ValueError, match=error_msg):
        DataColumn(sample_series, semantic_tags='time_index')

    data_col = DataColumn(sample_series)
    with pytest.raises(ValueError, match=error_msg):
        data_col.add_semantic_tags('time_index')
    with pytest.raises(ValueError, match=error_msg):
        data_col.set_semantic_tags('time_index')


def test_set_as_index(sample_series):
    data_col = DataColumn(sample_series)
    data_col._set_as_index()
    assert 'index' in data_col.semantic_tags


def test_set_as_time_index(sample_series):
    data_col = DataColumn(sample_series)
    data_col._set_as_time_index()
    assert 'time_index' in data_col.semantic_tags


def test_to_series(sample_series):
    data_col = DataColumn(sample_series)
    series = data_col.to_series()

    assert series is data_col._series
    pd.testing.assert_series_equal(to_pandas(series), to_pandas(data_col._series))


def test_shape(sample_series):
    col = DataColumn(sample_series)
    col_shape = col.shape
    series_shape = col.to_series().shape
    if dd and isinstance(sample_series, dd.Series):
        col_shape = (col_shape[0].compute(),)
        series_shape = (series_shape[0].compute(),)
    assert col_shape == (4,)
    assert col_shape == series_shape


def test_len(sample_series):
    col = DataColumn(sample_series)
    assert len(col) == len(sample_series) == 4


def test_dtype_update_on_init(datetime_series):
    dc = DataColumn(datetime_series,
                    logical_type='DateTime')
    assert dc._series.dtype == 'datetime64[ns]'


def test_dtype_update_on_ltype_change():
    dc = DataColumn(pd.Series([1, 2, 3]),
                    logical_type='Integer')
    assert dc._series.dtype == 'Int64'
    dc = dc.set_logical_type('Double')
    assert dc._series.dtype == 'float64'


def test_ordinal_requires_instance_on_init(sample_series):
    error_msg = 'Must use an Ordinal instance with order values defined'
    with pytest.raises(TypeError, match=error_msg):
        DataColumn(sample_series, logical_type=Ordinal)
    with pytest.raises(TypeError, match=error_msg):
        DataColumn(sample_series, logical_type="Ordinal")


def test_ordinal_requires_instance_on_update(sample_series):
    dc = DataColumn(sample_series, logical_type="NaturalLanguage")

    error_msg = 'Must use an Ordinal instance with order values defined'
    with pytest.raises(TypeError, match=error_msg):
        dc.set_logical_type(Ordinal)
    with pytest.raises(TypeError, match=error_msg):
        dc.set_logical_type("Ordinal")


def test_ordinal_with_order(sample_series):
    if (ks and isinstance(sample_series, ks.Series)) or (dd and isinstance(sample_series, dd.Series)):
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not compatible')

    ordinal_with_order = Ordinal(order=['a', 'b', 'c'])
    dc = DataColumn(sample_series, logical_type=ordinal_with_order)
    assert isinstance(dc.logical_type, Ordinal)
    assert dc.logical_type.order == ['a', 'b', 'c']

    dc = DataColumn(sample_series, logical_type="NaturalLanguage")
    new_dc = dc.set_logical_type(ordinal_with_order)
    assert isinstance(new_dc.logical_type, Ordinal)
    assert new_dc.logical_type.order == ['a', 'b', 'c']


def test_ordinal_with_incomplete_ranking(sample_series):
    if (ks and isinstance(sample_series, ks.Series)) or (dd and isinstance(sample_series, dd.Series)):
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not supported')

    ordinal_incomplete_order = Ordinal(order=['a', 'b'])
    error_msg = re.escape("Ordinal column sample_series contains values that are not "
                          "present in the order values provided: ['c']")
    with pytest.raises(ValueError, match=error_msg):
        DataColumn(sample_series, logical_type=ordinal_incomplete_order)


def test_ordinal_with_nan_values():
    nan_series = pd.Series(['a', 'b', np.nan, 'a'])
    ordinal_with_order = Ordinal(order=['a', 'b'])
    dc = DataColumn(nan_series, logical_type=ordinal_with_order)
    assert isinstance(dc.logical_type, Ordinal)
    assert dc.logical_type.order == ['a', 'b']


def test_datacolumn_equality(sample_series, sample_datetime_series):
    # Check different parameters to DataColumn
    str_col = DataColumn(sample_series, logical_type='Categorical')
    str_col_2 = DataColumn(sample_series, logical_type=Categorical)
    str_col_diff_tags = DataColumn(sample_series, logical_type=Categorical, semantic_tags={'test'})
    diff_name_col = DataColumn(sample_datetime_series, logical_type=Categorical)
    diff_dtype_col = DataColumn(sample_series, logical_type=NaturalLanguage)
    diff_description_col = DataColumn(sample_series, logical_type='Categorical', description='description')

    assert str_col == str_col_2
    assert str_col != str_col_diff_tags
    assert str_col != diff_name_col
    assert str_col != diff_dtype_col
    assert str_col != diff_description_col

    # Check columns with same logical types but different parameters
    ordinal_ltype_1 = Ordinal(order=['a', 'b', 'c'])
    ordinal_ltype_2 = Ordinal(order=['b', 'a', 'c'])
    ordinal_col_1 = DataColumn(sample_series, logical_type=ordinal_ltype_1)
    ordinal_col_2 = DataColumn(sample_series, logical_type=ordinal_ltype_2)

    assert str_col != ordinal_col_1
    assert ordinal_col_1 != ordinal_col_2
    assert ordinal_col_1 == ordinal_col_1

    datetime_ltype_instantiated = Datetime(datetime_format='%Y-%m%d')
    datetime_col_format = DataColumn(sample_datetime_series, logical_type=datetime_ltype_instantiated)
    datetime_col_param = DataColumn(sample_datetime_series, logical_type=Datetime(datetime_format=None))
    datetime_col_instantiated = DataColumn(sample_datetime_series, logical_type=Datetime())
    datetime_col = DataColumn(sample_datetime_series, logical_type=Datetime)

    assert datetime_col != datetime_col_instantiated
    assert datetime_col_instantiated != datetime_col_format
    assert datetime_col_instantiated == datetime_col_param

    # Check different underlying series
    str_col = DataColumn(sample_series, logical_type='NaturalLanguage')
    changed_series = sample_series.copy().replace(to_replace='a', value='test')
    null_col = DataColumn(changed_series, logical_type='NaturalLanguage')

    # We only check underlying data for equality with pandas dataframes
    if isinstance(str_col.to_series(), pd.Series):
        assert str_col != null_col
    else:
        assert str_col == null_col



import subprocess


def test_list_logical_types():
    subprocess.check_output(['python', '-m', 'woodwork', 'list-logical-types'])


def test_list_semantic_tags():
    subprocess.check_output(['python', '-m', 'woodwork', 'list-semantic-tags'])



import pandas as pd

from woodwork.utils import _get_specified_ltype_params, camel_to_snake


class ClassNameDescriptor(object):
    """Descriptor to convert a class's name from camelcase to snakecase"""

    def __get__(self, instance, class_):
        return camel_to_snake(class_.__name__)


class LogicalTypeMetaClass(type):

    def __repr__(cls):
        return cls.__name__


class LogicalType(object, metaclass=LogicalTypeMetaClass):
    """Base class for all other Logical Types"""
    type_string = ClassNameDescriptor()
    pandas_dtype = 'string'
    backup_dtype = None
    standard_tags = {}

    def __eq__(self, other, deep=False):
        return isinstance(other, self.__class__) and _get_specified_ltype_params(other) == _get_specified_ltype_params(self)

    def __str__(self):
        return str(self.__class__)


class Boolean(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.

    Examples:
        .. code-block:: python

            [True, False, True]
            [0, 1, 1]
    """
    pandas_dtype = 'boolean'
    backup_dtype = 'bool'


class Categorical(LogicalType):
    """Represents Logical Types that contain unordered discrete values that fall
    into one of a set of possible values. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["red", "green", "blue"]
            ["produce", "dairy", "bakery"]
            [3, 1, 2]
    """
    pandas_dtype = 'category'
    backup_dtype = 'str'
    standard_tags = {'category'}

    def __init__(self, encoding=None):
        # encoding dict(str -> int)
        # user can specify the encoding to use downstream
        pass


class CountryCode(LogicalType):
    """Represents Logical Types that contain categorical information specifically
    used to represent countries. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["AUS", "USA", "UKR"]
            ["GB", "NZ", "DE"]
    """
    pandas_dtype = 'category'
    backup_dtype = 'str'
    standard_tags = {'category'}


class Datetime(LogicalType):
    """Represents Logical Types that contain date and time information.

    Args:
        datetime_format (str): Desired datetime format for data

    Examples:
        .. code-block:: python

            ["2020-09-10",
             "2020-01-10 00:00:00",
             "01/01/2000 08:30"]
    """
    pandas_dtype = 'datetime64[ns]'
    datetime_format = None

    def __init__(self, datetime_format=None):
        self.datetime_format = datetime_format


class Double(LogicalType):
    """Represents Logical Types that contain positive and negative numbers, some of
    which include a fractional component. Includes zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [1.2, 100.4, 3.5]
            [-15.34, 100, 58.3]
    """
    pandas_dtype = 'float64'
    standard_tags = {'numeric'}


class Integer(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, 0]
            [-54, 73, 11]
    """
    pandas_dtype = 'Int64'
    backup_dtype = 'int64'
    standard_tags = {'numeric'}


class EmailAddress(LogicalType):
    """Represents Logical Types that contain email address values.

    Examples:
        .. code-block:: python

            ["john.smith@example.com",
             "support@example.com",
             "team@example.com"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class Filepath(LogicalType):
    """Represents Logical Types that specify locations of directories and files
    in a file system.

    Examples:
        .. code-block:: python

            ["/usr/local/bin",
             "/Users/john.smith/dev/index.html",
             "/tmp"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class FullName(LogicalType):
    """Represents Logical Types that may contain first, middle and last names,
    including honorifics and suffixes.

    Examples:
        .. code-block:: python

            ["Mr. John Doe, Jr.",
             "Doe, Mrs. Jane",
             "James Brown"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class IPAddress(LogicalType):
    """Represents Logical Types that contain IP addresses, including both
    IPv4 and IPv6 addresses.

    Examples:
        .. code-block:: python

            ["172.16.254.1",
             "192.0.0.0",
             "2001:0db8:0000:0000:0000:ff00:0042:8329"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class LatLong(LogicalType):
    """Represents Logical Types that contain latitude and longitude values

    Examples:
        .. code-block:: python

            [(33.670914, -117.841501),
             (40.423599, -86.921162)),
             (-45.031705, 168.659506)]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class NaturalLanguage(LogicalType):
    """Represents Logical Types that contain text or characters representing
    natural human language

    Examples:
        .. code-block:: python

            ["This is a short sentence.",
             "I like to eat pizza!",
             "When will humans go to mars?"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class Ordinal(LogicalType):
    """Represents Logical Types that contain ordered discrete values.
    Has 'category' as a standard tag.

    Args:
        order (list or tuple): An list or tuple specifying the order of the ordinal
            values from low to high. The underlying series cannot contain values that
            are not present in the order values.

    Examples:
        .. code-block:: python

            ["first", "second", "third"]
            ["bronze", "silver", "gold"]
    """
    pandas_dtype = 'category'
    backup_dtype = 'str'
    standard_tags = {'category'}

    def __init__(self, order):
        if not isinstance(order, (list, tuple)):
            raise TypeError("Order values must be specified in a list or tuple")
        if len(order) != len(set(order)):
            raise ValueError("Order values cannot contain duplicates")
        self.order = order

    def _validate_data(self, series):
        """Confirm the supplied series does not contain any values that are not
        in the specified order values"""
        if isinstance(series, pd.Series):
            missing_order_vals = set(series.dropna().values).difference(self.order)
            if missing_order_vals:
                error_msg = f'Ordinal column {series.name} contains values that are not present ' \
                    f'in the order values provided: {sorted(list(missing_order_vals))}'
                raise ValueError(error_msg)


class PhoneNumber(LogicalType):
    """Represents Logical Types that contain numeric digits and characters
    representing a phone number

    Examples:
        .. code-block:: python

            ["1-(555)-123-5495",
             "+1-555-123-5495",
             "5551235495"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class SubRegionCode(LogicalType):
    """Represents Logical Types that contain codes representing a portion of
    a larger geographic region. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["US-CO", "US-MA", "US-CA"]
            ["AU-NSW", "AU-TAS", "AU-QLD"]
    """
    pandas_dtype = 'category'
    backup_dtype = 'str'
    standard_tags = {'category'}


class Timedelta(LogicalType):
    """Represents Logical Types that contain values specifying a duration of time

    Examples:
        .. code-block:: python

            [pd.Timedelta('1 days 00:00:00'),
             pd.Timedelta('-1 days +23:40:00'),
             pd.Timedelta('4 days 12:00:00')]
    """
    pandas_dtype = 'timedelta64[ns]'


class URL(LogicalType):
    """Represents Logical Types that contain URLs, which may include protocol, hostname
    and file name

    Examples:
        .. code-block:: python

            ["http://google.com",
             "https://example.com/index.html",
             "example.com"]
    """
    pandas_dtype = 'string'
    backup_dtype = 'str'


class ZIPCode(LogicalType):
    """Represents Logical Types that contain a series of postal codes used by
    the US Postal Service for representing a group of addresses.
    Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["90210"
             "60018-0123",
             "10021"]
    """
    pandas_dtype = 'category'
    backup_dtype = 'str'
    standard_tags = {'category'}


def get_logical_types():
    """Returns a dictionary of logical type name strings and logical type classes"""
    # Get snake case strings
    logical_types = {logical_type.type_string: logical_type for logical_type in LogicalType.__subclasses__()}
    # Add class name strings
    class_name_dict = {logical_type.__name__: logical_type for logical_type in LogicalType.__subclasses__()}
    logical_types.update(class_name_dict)

    return logical_types


def str_to_logical_type(logical_str, params=None, raise_error=True):
    """Helper function for converting a string value to the corresponding logical type object.
    If a dictionary of params for the logical type is provided, apply them."""
    logical_str = logical_str.lower()
    logical_types_dict = {ltype_name.lower(): ltype for ltype_name, ltype in get_logical_types().items()}

    if logical_str in logical_types_dict:
        ltype = logical_types_dict[logical_str]
        if params:
            return ltype(**params)
        else:
            return ltype
    elif raise_error:
        raise ValueError('String %s is not a valid logical type' % logical_str)



import pytest

from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    FullName,
    LogicalType,
    Ordinal,
    get_logical_types,
    str_to_logical_type
)


def test_logical_eq():
    assert Boolean == Boolean
    assert Boolean() == Boolean()
    assert Categorical != Boolean
    assert Datetime != Datetime()
    assert Datetime() == Datetime(datetime_format=None)
    assert Datetime() != Datetime(datetime_format='%Y-%m-%d')


def test_logical_repr():
    assert isinstance(repr(Boolean), str)
    assert repr(Boolean) == 'Boolean'
    assert isinstance(repr(Categorical), str)
    assert repr(Categorical) == 'Categorical'


def test_instantiated_type_str():
    assert str(Categorical()) == 'Categorical'
    assert str(Boolean()) == 'Boolean'


def test_get_logical_types():
    all_types = LogicalType.__subclasses__()
    logical_types = get_logical_types()

    for logical_type in all_types:
        assert logical_types[logical_type.__name__] == logical_type
        assert logical_types[logical_type.type_string] == logical_type

    assert len(logical_types) == 2 * len(all_types)


def test_str_to_logical_type():
    all_types = LogicalType.__subclasses__()

    with pytest.raises(ValueError, match='String test is not a valid logical type'):
        str_to_logical_type('test')
    assert str_to_logical_type('test', raise_error=False) is None

    for logical_type in all_types:
        assert str_to_logical_type(logical_type.__name__) == logical_type
        assert str_to_logical_type(logical_type.type_string) == logical_type

    assert str_to_logical_type('bOoLeAn') == Boolean
    assert str_to_logical_type('full_NAME') == FullName
    assert str_to_logical_type('FullnamE') == FullName

    ymd = '%Y-%m-%d'
    datetime_with_format = str_to_logical_type('datetime', params={'datetime_format': ymd})
    assert datetime_with_format.__class__ == Datetime
    assert datetime_with_format.datetime_format == ymd
    assert datetime_with_format == Datetime(datetime_format=ymd)

    datetime_no_format = str_to_logical_type('datetime', params={'datetime_format': None})
    assert datetime_no_format.__class__ == Datetime
    assert datetime_no_format.datetime_format is None
    assert datetime_no_format == Datetime()

    # When parameters are supplied in a non-empty dictionary, the logical type gets instantiated
    assert str_to_logical_type('full_NAME', params={}) == FullName
    assert datetime_no_format != Datetime


def test_ordinal_order_errors():
    with pytest.raises(TypeError, match='Order values must be specified in a list or tuple'):
        Ordinal(order='not_valid')

    with pytest.raises(ValueError, match='Order values cannot contain duplicates'):
        Ordinal(order=['a', 'b', 'b'])


def test_ordinal_init_with_order():
    order = ['bronze', 'silver', 'gold']
    ordinal_from_list = Ordinal(order=order)
    assert ordinal_from_list.order == order

    order = ('bronze', 'silver', 'gold')
    ordinal_from_tuple = Ordinal(order=order)
    assert ordinal_from_tuple.order == order



import warnings

import numpy as np
import pandas as pd
from sklearn.metrics.cluster import normalized_mutual_info_score

import woodwork.serialize as serialize
from woodwork.datacolumn import DataColumn
from woodwork.exceptions import ColumnNameMismatchWarning
from woodwork.indexers import _iLocIndexer
from woodwork.logical_types import (
    Boolean,
    Datetime,
    Double,
    LogicalType,
    str_to_logical_type
)
from woodwork.utils import (
    _convert_input_to_set,
    _get_ltype_class,
    _get_mode,
    _is_numeric_series,
    _new_dt_including,
    col_is_datetime,
    import_or_none
)

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')
if ks:
    ks.set_option('compute.ops_on_diff_frames', True)


class DataTable(object):
    def __init__(self, dataframe,
                 name=None,
                 index=None,
                 time_index=None,
                 semantic_tags=None,
                 logical_types=None,
                 metadata=None,
                 use_standard_tags=True,
                 make_index=False,
                 column_descriptions=None,
                 already_sorted=False):
        """Create DataTable

        Args:
            dataframe (pd.DataFrame, dd.DataFrame, ks.DataFrame, numpy.ndarray): Dataframe providing the data for the datatable.
            name (str, optional): Name used to identify the datatable.
            index (str, optional): Name of the index column in the dataframe.
            time_index (str, optional): Name of the time index column in the dataframe.
            semantic_tags (dict, optional): Dictionary mapping column names in the dataframe to the
                semantic tags for the column. The keys in the dictionary should be strings
                that correspond to columns in the underlying dataframe. There are two options for
                specifying the dictionary values:
                (str): If only one semantic tag is being set, a single string can be used as a value.
                (list[str] or set[str]): If multiple tags are being set, a list or set of strings can be
                used as the value.
                Semantic tags will be set to an empty set for any column not included in the
                dictionary.
            logical_types (dict[str -> LogicalType], optional): Dictionary mapping column names in
                the dataframe to the LogicalType for the column. LogicalTypes will be inferred
                for any columns not present in the dictionary.
            metadata (dict[str -> json serializable], optional): Dictionary containing extra metadata for the DataTable.
            use_standard_tags (bool, optional): If True, will add standard semantic tags to columns based
                on the inferred or specified logical type for the column. Defaults to True.
            make_index (bool, optional): If True, will create a new unique, numeric index column with the
                name specified by ``index`` and will add the new index column to the supplied DataFrame.
                If True, the name specified in ``index`` cannot match an existing column name in
                ``dataframe``. If False, the name is specified in ``index`` must match a column
                present in the ``dataframe``. Defaults to False.
            column_descriptions (dict[str -> str], optional): Dictionary containing column descriptions
            already_sorted (bool, optional): Indicates whether the input dataframe is already sorted on the time
                index. If False, will sort the dataframe first on the time_index and then on the index (pandas DataFrame
                only). Defaults to False.
        """
        # Check that inputs are valid
        dataframe = _validate_dataframe(dataframe)
        _validate_params(dataframe, name, index, time_index, logical_types,
                         metadata, semantic_tags, make_index, column_descriptions)

        self._dataframe = dataframe

        self.make_index = make_index or None
        if self.make_index:
            self._dataframe = _make_index(self._dataframe, index)

        self.name = name
        self.use_standard_tags = use_standard_tags

        # Infer logical types and create columns
        self.columns = self._create_columns(self._dataframe.columns,
                                            logical_types,
                                            semantic_tags,
                                            use_standard_tags,
                                            column_descriptions)
        if index:
            _update_index(self, index)

        # Update dtypes before setting time index so that any Datetime formatting is applied
        self._update_columns(self.columns)

        if time_index:
            _update_time_index(self, time_index)
            self._sort_columns(already_sorted)

        self.metadata = metadata or {}

    def __eq__(self, other, deep=True):
        if self.name != other.name:
            return False
        if self.index != other.index:
            return False
        if self.time_index != other.time_index:
            return False
        if set(self.columns.keys()) != set(other.columns.keys()):
            return False
        for col_name in self.columns.keys():
            if self[col_name] != other[col_name]:
                return False

        # Only check pandas DataFrames for equality
        if isinstance(self.to_dataframe(), pd.DataFrame) and isinstance(other.to_dataframe(), pd.DataFrame):
            return self.to_dataframe().equals(other.to_dataframe())
        return True

    def __getitem__(self, key):
        if isinstance(key, list):
            if not all([isinstance(col, str) for col in key]):
                raise KeyError('Column names must be strings')
            invalid_cols = set(key).difference(set(self.columns.keys()))
            if invalid_cols:
                raise KeyError(f"Column(s) '{', '.join(sorted(list(invalid_cols)))}' not found in DataTable")
            return self._new_dt_from_cols(key)
        if not isinstance(key, str):
            raise KeyError('Column name must be a string')
        if key not in self.columns.keys():
            raise KeyError(f"Column with name '{key}' not found in DataTable")
        return self.columns[key]

    def __setitem__(self, col_name, column):
        if not isinstance(col_name, str):
            raise KeyError('Column name must be a string')

        if not isinstance(column, DataColumn):
            raise ValueError('New column must be of DataColumn type')

        # Don't allow reassigning of index or time index with setitem
        if self.index == col_name:
            raise KeyError('Cannot reassign index. Change column name and then use dt.set_index to reassign index.')
        if self.time_index == col_name:
            raise KeyError('Cannot reassign time index. Change column name and then use dt.set_time_index to reassign time index.')

        if column.name is not None and column.name != col_name:
            warnings.warn(ColumnNameMismatchWarning().get_warning_message(column.name, col_name),
                          ColumnNameMismatchWarning)
            column._series.name = col_name
            column._assigned_name = col_name

        self._dataframe[col_name] = column._series
        self._update_columns({col_name: column})

    def __sizeof__(self):
        return self._dataframe.__sizeof__()

    def __len__(self):
        return self._dataframe.__len__()

    @property
    def types(self):
        """Dataframe containing the physical dtypes, logical types and semantic
        tags for the table"""
        typing_info = {}
        # Access column names from underlying data to maintain column order
        for col_name in self._dataframe.columns:
            dc = self[col_name]
            typing_info[dc.name] = [dc.dtype, dc.logical_type, dc.semantic_tags]
        df = pd.DataFrame.from_dict(typing_info,
                                    orient='index',
                                    columns=['Physical Type', 'Logical Type', 'Semantic Tag(s)'],
                                    dtype="object")
        df.index.name = 'Data Column'
        return df

    @property
    def ltypes(self):
        """A series listing the logical types for each column in the table"""
        return self.types['Logical Type']

    def _create_columns(self,
                        column_names,
                        logical_types,
                        semantic_tags,
                        use_standard_tags,
                        column_descriptions):
        """Create a dictionary with column names as keys and new DataColumn objects
        as values, while assigning any values that are passed for logical types or
        semantic tags to the new column."""
        datacolumns = {}
        for name in column_names:
            if logical_types and name in logical_types:
                logical_type = logical_types[name]
            else:
                logical_type = None
            if semantic_tags and name in semantic_tags:
                semantic_tag = semantic_tags[name]
            else:
                semantic_tag = None
            if column_descriptions:
                description = column_descriptions.get(name)
            else:
                description = None
            dc = DataColumn(self._dataframe[name], logical_type, semantic_tag, use_standard_tags, name, description)
            datacolumns[dc.name] = dc
        return datacolumns

    @property
    def logical_types(self):
        """A dictionary containing logical types for each column"""
        return {dc.name: dc.logical_type for dc in self.columns.values()}

    @property
    def physical_types(self):
        """A dictionary containing physical types for each column"""
        return {dc.name: dc.dtype for dc in self.columns.values()}

    @property
    def semantic_tags(self):
        """A dictionary containing semantic tags for each column"""
        return {dc.name: dc.semantic_tags for dc in self.columns.values()}

    @property
    def shape(self):
        """Returns a tuple representing the dimensionality of the DataTable. If Dask DataFrame, returns
            a Dask `Delayed` object for the number of rows."""
        return self._dataframe.shape

    @property
    def index(self):
        """The index column for the table"""
        for column in self.columns.values():
            if 'index' in column.semantic_tags:
                return column.name
        return None

    @index.setter
    def index(self, index):
        if self.index and index is None:
            updated_index_col = self.columns[self.index].remove_semantic_tags('index')
            self._update_columns({self.index: updated_index_col})
        elif index is not None:
            _update_index(self, index, self.index)

    @property
    def time_index(self):
        """The time index column for the table"""
        for column in self.columns.values():
            if 'time_index' in column.semantic_tags:
                return column.name
        return None

    @time_index.setter
    def time_index(self, time_index):
        if self.time_index and time_index is None:
            updated_time_index_col = self.columns[self.time_index].remove_semantic_tags('time_index')
            self._update_columns({self.time_index: updated_time_index_col})
        elif time_index is not None:
            _update_time_index(self, time_index, self.time_index)

    def _update_columns(self, new_columns):
        """Update the DataTable columns based on items contained in the
        provided new_columns dictionary"""
        for name, column in new_columns.items():
            self.columns[name] = column
            # Make sure the underlying dataframe is in sync in case series data has changed
            self._dataframe[name] = column._series

    def _sort_columns(self, already_sorted):
        if dd and isinstance(self._dataframe, dd.DataFrame) or (ks and isinstance(self._dataframe, ks.DataFrame)):
            already_sorted = True  # Skip sorting for Dask and Koalas input
        if not already_sorted:
            sort_cols = [self.time_index, self.index]
            if not self.index:
                sort_cols = [self.time_index]
            self._dataframe = self._dataframe.sort_values(sort_cols)

    def pop(self, column_name):
        """Return a DataColumn and drop it from the DataTable.

        Args:
            column (str): Name of the column to pop.

        Returns:
            woodwork.DataColumn: DataColumn including logical type and semantic tags.
        """
        col = self[column_name]
        del self.columns[column_name]
        self._dataframe = self._dataframe.drop(column_name, axis=1)
        return col

    def rename(self, columns):
        """Renames columns in a DataTable

        Args:
            columns (dict[str -> str]): A dictionary mapping columns whose names
                we'd like to change to the name to which we'd like to change them.

        Returns:
            woodwork.DataTable: DataTable with the specified columns renamed.

        Note:
            Index and time index columns cannot be renamed.
        """
        if len(columns) != len(set(columns.values())):
            raise ValueError('New columns names must be unique from one another.')

        for old_name, new_name in columns.items():
            if not isinstance(old_name, str):
                raise KeyError(f"Column to rename must be a string. {old_name} is not a string.")
            if not isinstance(new_name, str):
                raise ValueError(f"New column name must be a string. {new_name} is not a string.")
            if old_name not in self.columns:
                raise KeyError(f"Column to rename must be present in the DataTable. {old_name} is not present in the DataTable.")
            if new_name in self.columns and new_name not in columns.keys():
                raise ValueError(f"The column {new_name} is already present in the DataTable. Please choose another name to rename {old_name} to or also rename {old_name}.")
            if old_name == self.index or old_name == self.time_index:
                raise KeyError(f"Cannot rename index or time index columns such as {old_name}.")

        old_all_cols = list(self._dataframe.columns)
        new_dt = self._new_dt_from_cols(old_all_cols)
        updated_cols = {}
        for old_name, new_name in columns.items():
            col = new_dt.pop(old_name)

            col._series.name = new_name
            col._assigned_name = new_name
            updated_cols[new_name] = col

        new_dt._update_columns(updated_cols)
        # Make sure we maintain relative order of columns including new names
        new_all_cols = [name if name not in columns else columns[name] for name in old_all_cols]
        new_dt._dataframe = new_dt._dataframe[new_all_cols]

        return new_dt

    def set_index(self, index):
        """Set the index column and return a new DataTable. Adds the 'index' semantic
        tag to the column and clears the tag from any previously set index column.
        Setting a column as the index column will also cause any previously set standard
        tags for the column to be removed.

        Args:
            index (str): The name of the column to set as the index

        Returns:
            woodwork.DataTable: DataTable with the specified index column set.
        """
        new_dt = self._new_dt_from_cols(self._dataframe.columns)
        _update_index(new_dt, index, self.index)
        return new_dt

    def set_time_index(self, time_index):
        """Set the time index column. Adds the 'time_index' semantic tag to the column and
        clears the tag from any previously set index column

        Args:
            time_index (str): The name of the column to set as the time index.
        """
        new_dt = self._new_dt_from_cols(self._dataframe.columns)
        _update_time_index(new_dt, time_index, self.time_index)
        return new_dt

    def set_types(self, logical_types=None, semantic_tags=None, retain_index_tags=True):
        """Update the logical type and semantic tags for any columns names in the provided types
        dictionary. Replaces existing columns with new DataColumn objects and returns a new
        DataTable object.

        Args:
            logical_types (dict[str -> str], optional): A dictionary defining the new logical types for the
                specified columns.
            semantic_tags (dict[str -> str/list/set], optional): A dictionary defining the new semantic_tags for the
                specified columns.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will replace all semantic tags. Defaults to
                True.

        Returns:
            woodwork.DataTable: DataTable with updated logical types and specified semantic tags set.
        """
        logical_types = logical_types or {}
        _check_logical_types(self._dataframe, logical_types)

        semantic_tags = semantic_tags or {}
        _check_semantic_tags(self._dataframe, semantic_tags)

        new_dt = self._new_dt_from_cols(self._dataframe.columns)
        cols_to_update = {}
        for col_name, col in new_dt.columns.items():
            if col_name not in logical_types and col_name not in semantic_tags:
                continue
            if col_name in logical_types:
                col = col.set_logical_type(logical_types[col_name], retain_index_tags)
            if col_name in semantic_tags:
                col = col.set_semantic_tags(semantic_tags[col_name], retain_index_tags)
            cols_to_update[col_name] = col

        new_dt._update_columns(cols_to_update)
        return new_dt

    def add_semantic_tags(self, semantic_tags):
        """Adds specified semantic tags to columns. Will retain any previously set values.
        Replaces updated columns with new DataColumn objects and returns a new DataTable object.

        Args:
            semantic_tags (dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataTable to the tags that should be added to the column

        Returns:
            woodwork.DataTable: DataTable with semantic tags added
        """
        _check_semantic_tags(self._dataframe, semantic_tags)
        return self._update_cols_and_get_new_dt('add_semantic_tags', semantic_tags)

    def remove_semantic_tags(self, semantic_tags):
        """Remove the semantic tags for any column names in the provided semantic_tags
        dictionary. Replaces the column with a new DataColumn object and return a new DataTable
        object.

        Args:
            semantic_tags (dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataTable to the tags that should be removed to the column

        Returns:
            woodwork.DataTable: DataTable with the specified semantic tags removed
        """
        _check_semantic_tags(self._dataframe, semantic_tags)
        return self._update_cols_and_get_new_dt('remove_semantic_tags', semantic_tags)

    def reset_semantic_tags(self, columns=None, retain_index_tags=False):
        """Reset the semantic tags for the specified columns to the default values and
        return a new DataTable. The default values will be either an empty set or a set
        of the standard tags based on the column logical type, controlled by the
        use_standard_tags property on the table. Columns names can be provided as a
        single string, a list of strings or a set of strings. If columns is not specified,
        tags will be reset for all columns.

        Args:
            columns (str/list/set): The columns for which the semantic tags should be reset.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will clear all semantic tags. Defaults to
                False.

        Returns:
            woodwork.DataTable: DataTable with semantic tags reset to default values
        """
        columns = _convert_input_to_set(columns, "columns")
        cols_not_found = sorted(list(columns.difference(set(self._dataframe.columns))))
        if cols_not_found:
            raise LookupError("Input contains columns that are not present in "
                              f"dataframe: '{', '.join(cols_not_found)}'")
        if not columns:
            columns = self._dataframe.columns
        return self._update_cols_and_get_new_dt('reset_semantic_tags', columns, retain_index_tags)

    def _update_cols_and_get_new_dt(self, method, new_values, *args):
        """Helper method that can be used for updating columns by calling the column method
        that is specified, passing along information contained in new_values and any
        additional positional arguments.

        Args:
            method (str): The name of the method to call on the DataColumn object to perform the update.
            new_values (dict/list): If a dictionary is provided the keys should correspond to column
                names and the items in the dictionary values will be passed along to the DataColumn method.
                If a list is provided, the items in the list should correspond to column names and
                no additional values will be passed along to the DataColumn method.

        Returns:
            woodwork.DataTable: A new DataTable with updated columns
        """
        new_dt = self._new_dt_from_cols(self._dataframe.columns)
        cols_to_update = {}
        if isinstance(new_values, dict):
            for name, tags in new_values.items():
                cols_to_update[name] = getattr(new_dt.columns[name], method)(tags, *args)
        else:
            for name in new_values:
                cols_to_update[name] = getattr(new_dt.columns[name], method)(*args)
        new_dt._update_columns(cols_to_update)
        return new_dt

    def to_dataframe(self):
        """Retrieves the DataTable's underlying dataframe.

        Note: Do not modify the returned dataframe directly to avoid unexpected behavior

        Returns:
            DataFrame: The underlying dataframe of the DataTable. Return type will depend on the type
                of dataframe used to create the DataTable.
        """
        if self.index is None or not isinstance(self._dataframe, pd.DataFrame):
            return self._dataframe
        else:
            return self._dataframe.set_index(self.index, drop=False)

    def select(self, include):
        """Create a DataTable including only columns whose logical type and
        semantic tags are specified in the list of types and tags to include.
        If no matching columns are found, an empty DataTable will be returned.

        Args:
            include (str or LogicalType or list[str or LogicalType]): Logical
                types, semantic tags or column names to include
                in the DataTable.

        Returns:
            DataTable: The subset of the original DataTable that contains just the
            logical types and semantic tags in ``include``.
        """
        cols_to_include = self._filter_cols(include)
        return self._new_dt_from_cols(cols_to_include)

    def update_dataframe(self, new_df, already_sorted=False):
        '''Replace the DataTable's dataframe with a new dataframe, making sure the new dataframe dtypes are updated.
        If the original DataTable was created with ``make_index=True``, an index column will be added to the updated
        data if it is not present.

        Args:
            new_df (DataFrame): Dataframe containing the new data. The same columns present in the original data should
                also be present in the new dataframe.
            already_sorted (bool, optional): Indicates whether the input dataframe is already sorted on the time
                index. If False, will sort the dataframe first on the time_index and then on the index (pandas DataFrame
                only). Defaults to False.
        '''
        if self.make_index and self.index not in new_df.columns:
            new_df = _make_index(new_df, self.index)

        if len(new_df.columns) != len(self.columns):
            raise ValueError("Updated dataframe contains {} columns, expecting {}".format(len(new_df.columns),
                                                                                          len(self.columns)))
        for column in self.columns.keys():
            if column not in new_df.columns:
                raise ValueError("Updated dataframe is missing new {} column".format(column))

        if self.index:
            _check_index(new_df, self.index)

        # Make sure column ordering matches existing ordering
        new_df = new_df[[column for column in self._dataframe.columns]]
        self._dataframe = new_df

        if self.time_index:
            _check_time_index(new_df, self.time_index)
            self._sort_columns(already_sorted)

        # Update column series and dtype
        for column in self.columns.keys():
            self.columns[column]._set_series(new_df[column])
            self.columns[column]._update_dtype()

        # Make sure dataframe dtypes match columns
        self._update_columns(self.columns)

    def _filter_cols(self, include, col_names=False):
        """Return list of columns filtered in specified way. In case of collision, favors logical types
        then semantic tag then column name.

        Args:
            include (str or LogicalType or list[str or LogicalType]): parameter or list of parameters to
                filter columns by.

            col_names (bool): Specifies whether to filter columns by name. Defaults to False.

        Returns:
            List[str] of column names that fit into filter.
        """
        if not isinstance(include, list):
            include = [include]

        ltypes_used = set()
        ltypes_in_dt = {_get_ltype_class(col.logical_type) for col in self.columns.values()}

        tags_used = set()
        tags_in_dt = {tag for col in self.columns.values() for tag in col.semantic_tags}

        cols_to_include = set()

        for selector in include:
            if _get_ltype_class(selector) in LogicalType.__subclasses__():
                if selector not in LogicalType.__subclasses__():
                    raise TypeError(f"Invalid selector used in include: {selector} cannot be instantiated")
                if selector in ltypes_in_dt:
                    ltypes_used.add(selector)
            elif isinstance(selector, str):
                # If the str is a viable ltype, it'll take precedence
                # but if it's not present, we'll check if it's a tag
                ltype = str_to_logical_type(selector, raise_error=False)
                if ltype and ltype in ltypes_in_dt:
                    ltypes_used.add(ltype)
                    continue
                elif selector in tags_in_dt:
                    tags_used.add(selector)
                elif col_names and selector in self.columns:
                    cols_to_include.add(selector)
            else:
                raise TypeError(f"Invalid selector used in include: {selector} must be either a string or LogicalType")

        for col_name, col in self.columns.items():
            if _get_ltype_class(col.logical_type) in ltypes_used or col.semantic_tags.intersection(tags_used):
                cols_to_include.add(col_name)

        # Maintain column order by using underlying data
        return [col_name for col_name in self._dataframe.columns if col_name in cols_to_include]

    def _new_dt_from_cols(self, cols_to_include):
        """Creates a new DataTable from a list of column names, retaining all types,
        indices, and name of original DataTable. Resulting DataTable's column order will
        follow the order used in cols_to_include."""
        assert all([col_name in self.columns for col_name in cols_to_include])
        return _new_dt_including(self, self._dataframe.loc[:, cols_to_include])

    @property
    def iloc(self):
        """Purely integer-location based indexing for selection by position.
        ``.iloc[]`` is primarily integer position based (from ``0`` to
        ``length-1`` of the axis), but may also be used with a boolean array.

        Allowed inputs are:
            An integer, e.g. ``5``.
            A list or array of integers, e.g. ``[4, 3, 0]``.
            A slice object with ints, e.g. ``1:7``.
            A boolean array.
            A ``callable`` function with one argument (the calling Series, DataFrame
            or Panel) and that returns valid output for indexing (one of the above).
            This is useful in method chains, when you don't have a reference to the
            calling object, but would like to base your selection on some value.
        """
        return _iLocIndexer(self)

    def describe_dict(self, include=None):
        """Calculates statistics for data contained in DataTable.

        Args:
            include (list[str or LogicalType], optional): filter for what columns to include in the
            statistics returned. Can be a list of columns, semantic tags, logical types, or a list
            combining any of the three. It follows the most broad specification. Favors logical types
            then semantic tag then column name. If no matching columns are found, an empty DataFrame
            will be returned.

        Returns:
            dict[str -> dict]: A dictionary with a key for each column in the data or for each column
            matching the logical types, semantic tags or column names specified in ``include``, paired
            with a value containing a dictionary containing relevant statistics for that column.
        """
        agg_stats_to_calculate = {
            'category': ["count", "nunique"],
            'numeric': ["count", "max", "min", "nunique", "mean", "std"],
            Datetime: ["count", "max", "min", "nunique", "mean"],
        }
        if include is not None:
            filtered_cols = self._filter_cols(include, col_names=True)
            cols_to_include = [(k, v) for k, v in self.columns.items() if k in filtered_cols]
        else:
            cols_to_include = self.columns.items()

        results = {}

        if dd and isinstance(self._dataframe, dd.DataFrame):
            df = self._dataframe.compute()
        elif ks and isinstance(self._dataframe, ks.DataFrame):
            # Missing values in Koalas will be replaced with 'None' - change them to
            # np.nan so stats are calculated properly
            df = self._dataframe.to_pandas().replace(to_replace='None', value=np.nan)
        else:
            df = self._dataframe

        for column_name, column in cols_to_include:
            if 'index' in column.semantic_tags:
                continue
            values = {}
            logical_type = column.logical_type
            semantic_tags = column.semantic_tags
            series = df[column_name]

            # Calculate Aggregation Stats
            if column._is_categorical():
                agg_stats = agg_stats_to_calculate['category']
            elif column._is_numeric():
                agg_stats = agg_stats_to_calculate['numeric']
            elif _get_ltype_class(logical_type) == Datetime:
                agg_stats = agg_stats_to_calculate[Datetime]
            else:
                agg_stats = ["count"]
            values = series.agg(agg_stats).to_dict()

            # Calculate other specific stats based on logical type or semantic tags
            if _get_ltype_class(logical_type) == Boolean:
                values["num_false"] = series.value_counts().get(False, 0)
                values["num_true"] = series.value_counts().get(True, 0)
            elif column._is_numeric():
                quant_values = series.quantile([0.25, 0.5, 0.75]).tolist()
                values["first_quartile"] = quant_values[0]
                values["second_quartile"] = quant_values[1]
                values["third_quartile"] = quant_values[2]

            values["nan_count"] = series.isna().sum()
            values["mode"] = _get_mode(series)
            values["physical_type"] = column.dtype
            values["logical_type"] = logical_type
            values["semantic_tags"] = semantic_tags
            results[column_name] = values
        return results

    def describe(self, include=None):
        """Calculates statistics for data contained in DataTable.

        Args:
            include (list[str or LogicalType], optional): filter for what columns to include in the
            statistics returned. Can be a list of columns, semantic tags, logical types, or a list
            combining any of the three. It follows the most broad specification. Favors logical types
            then semantic tag then column name. If no matching columns are found, an empty DataFrame
            will be returned.

        Returns:
            pd.DataFrame: A Dataframe containing statistics for the data or the subset of the original
            DataTable that contains the logical types, semantic tags, or column names specified
            in ``include``.
        """
        results = self.describe_dict(include=include)
        index_order = [
            'physical_type',
            'logical_type',
            'semantic_tags',
            'count',
            'nunique',
            'nan_count',
            'mean',
            'mode',
            'std',
            'min',
            'first_quartile',
            'second_quartile',
            'third_quartile',
            'max',
            'num_true',
            'num_false',
        ]
        return pd.DataFrame(results).reindex(index_order)

    def value_counts(self, ascending=False, top_n=10, dropna=False):
        """Returns a list of dictionaries with counts for the most frequent values in each column (only
            for DataColumns with `category` as a standard tag).


        Args:
            ascending (bool): Defines whether each list of values should be sorted most frequent
                to least frequent value (False), or least frequent to most frequent value (True).
                Defaults to False.

            top_n (int): the number of top values to retrieve. Defaults to 10.

            dropna (bool): determines whether to remove NaN values when finding frequency. Defaults
                to False.

        Returns:
            top_list (list(dict)): a list of dictionaries for each categorical column with keys `count`
                and `value`.
        """
        val_counts = {}
        valid_cols = [col for col, column in self.columns.items() if column._is_categorical()]
        data = self._dataframe[valid_cols]
        is_ks = False
        if dd and isinstance(data, dd.DataFrame):
            data = data.compute()
        if ks and isinstance(data, ks.DataFrame):
            data = data.to_pandas()
            is_ks = True

        for col in valid_cols:
            if dropna and is_ks:
                # Koalas categorical columns will have missing values replaced with the string 'None'
                # Replace them with np.nan so dropna work
                datacol = data[col].replace(to_replace='None', value=np.nan)
            else:
                datacol = data[col]
            frequencies = datacol.value_counts(ascending=ascending, dropna=dropna)
            df = frequencies[:top_n].reset_index()
            df.columns = ["value", "count"]
            dt_list = list(df.to_dict(orient="index").values())
            val_counts[col] = dt_list
        return val_counts

    def _handle_nans_for_mutual_info(self, data):
        """
        Remove NaN values in the dataframe so that mutual information can be calculated

        Args:
            data (pd.DataFrame): dataframe to use for calculating mutual information

        Returns:
            pd.DataFrame: data with fully null columns removed and nans filled in
                with either mean or mode

        """
        # remove fully null columns
        data = data.loc[:, data.columns[data.notnull().any()]]

        # replace or remove null values
        for column_name in data.columns[data.isnull().any()]:
            column = self[column_name]
            series = data[column_name]
            ltype = column._logical_type

            if column._is_numeric():
                mean = series.mean()
                if isinstance(mean, float) and not _get_ltype_class(ltype) == Double:
                    data[column_name] = series.astype('float')
                data[column_name] = series.fillna(mean)
            elif column._is_categorical() or _get_ltype_class(ltype) == Boolean:
                mode = _get_mode(series)
                data[column_name] = series.fillna(mode)
        return data

    def _make_categorical_for_mutual_info(self, data, num_bins):
        """Transforms dataframe columns into numeric categories so that
        mutual information can be calculated

        Args:
            data (pd.DataFrame): dataframe to use for caculating mutual information
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical.


        Returns:
            data (pd.DataFrame): Transformed data
        """

        col_names = data.columns.to_list()
        for col_name in col_names:
            if self[col_name]._is_numeric():
                # bin numeric features to make categories
                data[col_name] = pd.qcut(data[col_name], num_bins, duplicates="drop")
            # Convert Datetimes to total seconds - an integer - and bin
            if _get_ltype_class(self[col_name].logical_type) == Datetime:
                data[col_name] = pd.qcut(data[col_name].astype('int64'), num_bins, duplicates="drop")
            # convert categories to integers
            new_col = data[col_name]
            if str(new_col.dtype) != 'category':
                new_col = new_col.astype('category')
            data[col_name] = new_col.cat.codes
        return data

    def mutual_information_dict(self, num_bins=10, nrows=None):
        """
        Calculates mutual information between all pairs of columns in the DataTable that
        support mutual information. Logical Types that support mutual information are as
        follows:  Boolean, Categorical, CountryCode, Datetime, Double, Integer, Ordinal,
        SubRegionCode, and ZIPCode

        Args:
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical.
            nrows (int): The number of rows to sample for when determining mutual info.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.

        Returns:
            list(dict): A list containing dictionaries that have keys `column_1`,
            `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
            Mutual information values are between 0 (no mutual information) and 1
            (perfect dependency).
        """
        # We only want Numeric, Categorical, and Boolean columns
        # And we don't want the index column
        valid_columns = [col_name for col_name, column
                         in self.columns.items() if (col_name != self.index and
                                                     (column._is_numeric() or
                                                      column._is_categorical() or
                                                      _get_ltype_class(column.logical_type) == Boolean or
                                                      _get_ltype_class(column.logical_type) == Datetime)
                                                     )]

        data = self._dataframe[valid_columns]
        if dd and isinstance(data, dd.DataFrame):
            data = data.compute()
        if ks and isinstance(self._dataframe, ks.DataFrame):
            data = data.to_pandas()

        # cut off data if necessary
        if nrows is not None and nrows < data.shape[0]:
            data = data.sample(nrows)

        data = self._handle_nans_for_mutual_info(data)
        data = self._make_categorical_for_mutual_info(data, num_bins)

        # calculate mutual info for all pairs of columns
        mutual_info = []
        col_names = data.columns.to_list()
        for i, a_col in enumerate(col_names):
            for j in range(i, len(col_names)):
                b_col = col_names[j]
                if a_col == b_col:
                    # Ignore because the mutual info for a column with itself will always be 1
                    continue
                else:
                    mi_score = normalized_mutual_info_score(data[a_col], data[b_col])
                    mutual_info.append(
                        {"column_1": a_col, "column_2": b_col, "mutual_info": mi_score}
                    )
        mutual_info.sort(key=lambda mi: mi['mutual_info'], reverse=True)
        return mutual_info

    def mutual_information(self, num_bins=10, nrows=None):
        """
        Calculates mutual information between all pairs of columns in the DataTable that
        support mutual information. Logical Types that support mutual information are as
        follows:  Boolean, Categorical, CountryCode, Datetime, Double, Integer, Ordinal,
        SubRegionCode, and ZIPCode

        Args:
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical.
            nrows (int): The number of rows to sample for when determining mutual info.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.

        Returns:
            pd.DataFrame: A Dataframe containing mutual information with columns `column_1`,
            `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
            Mutual information values are between 0 (no mutual information) and 1
            (perfect dependency).
        """
        mutual_info = self.mutual_information_dict(num_bins, nrows)
        return pd.DataFrame(mutual_info)

    def to_dictionary(self):
        '''
        Get a DataTable's description

        Returns:
            description (dict) : Description of :class:`.DataTable`.
        '''
        return serialize.datatable_to_description(self)

    def to_csv(self, path, sep=',', encoding='utf-8', engine='python', compression=None, profile_name=None):
        '''Write DataTable to disk in the CSV format, location specified by `path`.
            Path could be a local path or a S3 path.
            If writing to S3 a tar archive of files will be written.

            Args:
                path (str) : Location on disk to write to (will be created as a directory)
                sep (str) : String of length 1. Field delimiter for the output file.
                encoding (str) : A string representing the encoding to use in the output file, defaults to 'utf-8'.
                engine (str) : Name of the engine to use. Possible values are: {'c', 'python'}.
                compression (str) : Name of the compression to use. Possible values are: {'gzip', 'bz2', 'zip', 'xz', None}.
                profile_name (str) : Name of AWS profile to use, False to use an anonymous profile, or None.
        '''
        serialize.write_datatable(self, path, format='csv', index=False,
                                  sep=sep, encoding=encoding, engine=engine,
                                  compression=compression, profile_name=profile_name)

    def to_pickle(self, path, compression=None, profile_name=None):
        '''Write DataTable to disk in the pickle format, location specified by `path`.
            Path could be a local path or a S3 path.
            If writing to S3 a tar archive of files will be written.

            Args:
                path (str) : Location on disk to write to (will be created as a directory)
                compression (str) : Name of the compression to use. Possible values are: {'gzip', 'bz2', 'zip', 'xz', None}.
                profile_name (str) : Name of AWS profile to use, False to use an anonymous profile, or None.
        '''
        serialize.write_datatable(self, path, format='pickle',
                                  compression=compression, profile_name=profile_name)

    def to_parquet(self, path, compression=None, profile_name=None):
        '''Write DataTable to disk in the parquet format, location specified by `path`.
            Path could be a local path or a S3 path.
            If writing to S3 a tar archive of files will be written.

            Note:
                As the engine `fastparquet` cannot handle nullable pandas dtypes, `pyarrow` will be used
                for serialization to parquet.

            Args:
                path (str): location on disk to write to (will be created as a directory)
                compression (str) : Name of the compression to use. Possible values are: {'snappy', 'gzip', 'brotli', None}.
                profile_name (str) : Name of AWS profile to use, False to use an anonymous profile, or None.
        '''
        serialize.write_datatable(self, path, format='parquet',
                                  engine='pyarrow', compression=compression,
                                  profile_name=profile_name)


def _validate_dataframe(dataframe):
    '''Check that the dataframe supplied during DataTable initialization is valid,
    and convert numpy array to pandas DataFrame if necessary.'''
    if not ((dd and isinstance(dataframe, dd.DataFrame)) or
            (ks and isinstance(dataframe, ks.DataFrame)) or
            isinstance(dataframe, (pd.DataFrame, np.ndarray))):
        raise TypeError('Dataframe must be one of: pandas.DataFrame, dask.DataFrame, koalas.DataFrame, numpy.ndarray')

    if isinstance(dataframe, np.ndarray):
        dataframe = pd.DataFrame(dataframe)
        dataframe.columns = dataframe.columns.astype(str)
    return dataframe


def _validate_params(dataframe, name, index, time_index, logical_types,
                     metadata, semantic_tags, make_index, column_descriptions):
    """Check that values supplied during DataTable initialization are valid"""
    _check_unique_column_names(dataframe)
    if name and not isinstance(name, str):
        raise TypeError('DataTable name must be a string')
    if index or make_index:
        _check_index(dataframe, index, make_index)
    if logical_types:
        _check_logical_types(dataframe, logical_types)
    if metadata:
        _check_metadata(metadata)
    if time_index:
        datetime_format = None
        logical_type = None
        if logical_types is not None and time_index in logical_types:
            logical_type = logical_types[time_index]
            if _get_ltype_class(logical_types[time_index]) == Datetime:
                datetime_format = logical_types[time_index].datetime_format

        _check_time_index(dataframe, time_index, datetime_format=datetime_format, logical_type=logical_type)

    if semantic_tags:
        _check_semantic_tags(dataframe, semantic_tags)

    if column_descriptions:
        _check_column_descriptions(dataframe, column_descriptions)


def _check_unique_column_names(dataframe):
    if not dataframe.columns.is_unique:
        raise IndexError('Dataframe cannot contain duplicate columns names')


def _check_index(dataframe, index, make_index=False):
    if index and not isinstance(index, str):
        raise TypeError('Index column name must be a string')
    if not make_index and index not in dataframe.columns:
        # User specifies an index that is not in the dataframe, without setting make_index to True
        raise LookupError(f'Specified index column `{index}` not found in dataframe. To create a new index column, set make_index to True.')
    if index and not make_index and isinstance(dataframe, pd.DataFrame) and not dataframe[index].is_unique:
        # User specifies an index that is in the dataframe but not unique
        # Does not check for Dask as Dask does not support is_unique
        raise IndexError('Index column must be unique')
    if make_index and index and index in dataframe.columns:
        # User sets make_index to True, but supplies an index name that matches a column already present
        raise IndexError('When setting make_index to True, the name specified for index cannot match an existing column name')
    if make_index and not index:
        # User sets make_index to True, but does not supply a name for the index
        raise IndexError('When setting make_index to True, the name for the new index must be specified in the index parameter')


def _check_time_index(dataframe, time_index, datetime_format=None, logical_type=None):
    if not isinstance(time_index, str):
        raise TypeError('Time index column name must be a string')
    if time_index not in dataframe.columns:
        raise LookupError(f'Specified time index column `{time_index}` not found in dataframe')
    if not (_is_numeric_series(dataframe[time_index], logical_type) or
            col_is_datetime(dataframe[time_index], datetime_format=datetime_format)):
        raise TypeError('Time index column must contain datetime or numeric values')


def _check_logical_types(dataframe, logical_types):
    if not isinstance(logical_types, dict):
        raise TypeError('logical_types must be a dictionary')
    cols_not_found = set(logical_types.keys()).difference(set(dataframe.columns))
    if cols_not_found:
        raise LookupError('logical_types contains columns that are not present in '
                          f'dataframe: {sorted(list(cols_not_found))}')


def _check_semantic_tags(dataframe, semantic_tags):
    if not isinstance(semantic_tags, dict):
        raise TypeError('semantic_tags must be a dictionary')
    cols_not_found = set(semantic_tags.keys()).difference(set(dataframe.columns))
    if cols_not_found:
        raise LookupError('semantic_tags contains columns that are not present in '
                          f'dataframe: {sorted(list(cols_not_found))}')


def _check_column_descriptions(dataframe, column_descriptions):
    if not isinstance(column_descriptions, dict):
        raise TypeError('column_descriptions must be a dictionary')
    cols_not_found = set(column_descriptions.keys()).difference(set(dataframe.columns))
    if cols_not_found:
        raise LookupError('column_descriptions contains columns that are not present in '
                          f'dataframe: {sorted(list(cols_not_found))}')


def _check_metadata(metadata):
    if not isinstance(metadata, dict):
        raise TypeError('Metadata must be a dictionary.')


def _update_index(datatable, index, old_index=None):
    """Add the `index` tag to the specified index column and remove the tag from the
    old_index column, if specified. Also checks that the specified index column
    can be used as an index."""
    _check_index(datatable._dataframe, index)
    datatable.columns[index]._set_as_index()
    if old_index:
        datatable._update_columns({old_index: datatable.columns[old_index].remove_semantic_tags('index')})


def _update_time_index(datatable, time_index, old_time_index=None):
    """Add the `time_index` tag to the specified time_index column and remove the tag from the
    old_time_index column, if specified. Also checks that the specified time_index
    column can be used as a time index."""

    _check_time_index(datatable._dataframe, time_index)
    datatable.columns[time_index]._set_as_time_index()
    if old_time_index:
        datatable._update_columns({old_time_index: datatable.columns[old_time_index].remove_semantic_tags('time_index')})


def _make_index(dataframe, index):
    if dd and isinstance(dataframe, dd.DataFrame):
        dataframe[index] = 1
        dataframe[index] = dataframe[index].cumsum() - 1
    elif ks and isinstance(dataframe, ks.DataFrame):
        dataframe = dataframe.koalas.attach_id_column('distributed-sequence', index)
    else:
        dataframe.insert(0, index, range(len(dataframe)))

    return dataframe




import numpy as np
import pandas as pd
import pytest


def pd_to_dask(series):
    dd = pytest.importorskip('dask.dataframe', reason='Dask not installed, skipping')
    return dd.from_pandas(series, npartitions=2)


def pd_to_koalas(series):
    ks = pytest.importorskip('databricks.koalas', reason='Koalas not installed, skipping')
    return ks.from_pandas(series)


# Integer Inference Fixtures
@pytest.fixture
def pandas_integers():
    return [
        pd.Series([-1, 2, 1, 7]),
        pd.Series([-1, 0, 5, 3]),
    ]


@pytest.fixture
def dask_integers(pandas_integers):
    return [pd_to_dask(series) for series in pandas_integers]


@pytest.fixture
def koalas_integers(pandas_integers):
    return [pd_to_koalas(series) for series in pandas_integers]


@pytest.fixture(params=['pandas_integers', 'dask_integers', 'koalas_integers'])
def integers(request):
    return request.getfixturevalue(request.param)


# Double Inference Fixtures
@pytest.fixture
def pandas_doubles():
    return [
        pd.Series([-1, 2.5, 1, 7]),
        pd.Series([1.5, np.nan, 1, 3])
    ]


@pytest.fixture
def dask_doubles(pandas_doubles):
    return [pd_to_dask(series) for series in pandas_doubles]


@pytest.fixture
def koalas_doubles(pandas_doubles):
    return [pd_to_koalas(series) for series in pandas_doubles]


@pytest.fixture(params=['pandas_doubles', 'dask_doubles', 'koalas_doubles'])
def doubles(request):
    return request.getfixturevalue(request.param)


# Boolean Inference Fixtures
@pytest.fixture
def pandas_bools():
    return [
        pd.Series([True, False, True, True]),
        pd.Series([True, np.nan, True, True]),
    ]


@pytest.fixture
def dask_bools(pandas_bools):
    return [pd_to_dask(series) for series in pandas_bools]


@pytest.fixture
def koalas_bools(pandas_bools):
    return [pd_to_koalas(series) for series in pandas_bools]


@pytest.fixture(params=['pandas_bools', 'dask_bools', 'koalas_bools'])
def bools(request):
    return request.getfixturevalue(request.param)


# Datetime Inference Fixtures
@pytest.fixture
def pandas_datetimes():
    return [
        pd.Series(['3/11/2000', '3/12/2000', '3/13/2000', '3/14/2000']),
        pd.Series(['3/11/2000', np.nan, '3/13/2000', '3/14/2000']),
    ]


@pytest.fixture
def dask_datetimes(pandas_datetimes):
    return [pd_to_dask(series) for series in pandas_datetimes]


@pytest.fixture
def koalas_datetimes(pandas_datetimes):
    return [pd_to_koalas(series) for series in pandas_datetimes]


@pytest.fixture(params=['pandas_datetimes', 'dask_datetimes', 'koalas_datetimes'])
def datetimes(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def datetime_series(datetimes):
    return datetimes[1]


# Categorical Inference Fixtures
@pytest.fixture
def pandas_categories():
    return [
        pd.Series(['a', 'b', 'a', 'b']),
        pd.Series(['1', '2', '1', '2']),
        pd.Series(['a', np.nan, 'b', 'b']),
        pd.Series([1, 2, 1, 2])
    ]


@pytest.fixture
def dask_categories(pandas_categories):
    return [pd_to_dask(series) for series in pandas_categories]


@pytest.fixture
def koalas_categories(pandas_categories):
    return [pd_to_koalas(series) for series in pandas_categories]


@pytest.fixture(params=['pandas_categories', 'dask_categories', 'koalas_categories'])
def categories(request):
    return request.getfixturevalue(request.param)


# Timedelta Inference Fixtures
@pytest.fixture
def pandas_timedeltas():
    return [
        pd.Series(pd.to_timedelta(range(4), unit='s')),
        pd.Series([pd.to_timedelta(1, unit='s'), np.nan])
    ]


@pytest.fixture
def dask_timedeltas(pandas_timedeltas):
    return [pd_to_dask(series) for series in pandas_timedeltas]


@pytest.fixture(params=['pandas_timedeltas', 'dask_timedeltas'])
def timedeltas(request):
    return request.getfixturevalue(request.param)


# NaturalLanguage Inference Fixtures
@pytest.fixture
def pandas_strings():
    return [
        pd.Series(['Mr. John Doe', 'Doe, Mrs. Jane', 'James Brown', 'Ms. Paige Turner']),
    ]


@pytest.fixture
def dask_strings(pandas_strings):
    return [pd_to_dask(series) for series in pandas_strings]


@pytest.fixture
def koalas_strings(pandas_strings):
    return [pd_to_koalas(series) for series in pandas_strings]


@pytest.fixture(params=['pandas_strings', 'dask_strings', 'koalas_strings'])
def strings(request):
    return request.getfixturevalue(request.param)


# NaturalLanguage Inference with Threshold
@pytest.fixture
def pandas_long_strings():
    natural_language_series = pd.Series([
        '01234567890123456789',
        '01234567890123456789',
        '01234567890123456789',
        '01234567890123456789'])
    category_series = pd.Series([
        '0123456789012345678',
        '0123456789012345678',
        '0123456789012345678',
        '0123456789012345678'])
    return [natural_language_series, category_series]


@pytest.fixture
def dask_long_strings(pandas_long_strings):
    return [pd_to_dask(series) for series in pandas_long_strings]


@pytest.fixture
def koalas_long_strings(pandas_long_strings):
    return [pd_to_koalas(series) for series in pandas_long_strings]


@pytest.fixture(params=['pandas_long_strings', 'dask_long_strings', 'koalas_long_strings'])
def long_strings(request):
    return request.getfixturevalue(request.param)


# pd.NA Inference Fixtures
@pytest.fixture
def pandas_pdnas():
    return [
        pd.Series(['Mr. John Doe', pd.NA, 'James Brown', 'Ms. Paige Turner']).astype('string'),
        pd.Series([1, pd.NA, 2, 3]).astype('Int64'),
        pd.Series([True, pd.NA, False, True]).astype('boolean'),
    ]


@pytest.fixture
def dask_pdnas(pandas_pdnas):
    return [pd_to_dask(series) for series in pandas_pdnas]


@pytest.fixture(params=['pandas_pdnas', 'dask_pdnas'])
def pdnas(request):
    return request.getfixturevalue(request.param)



import urllib.request

import pytest

from woodwork import DataTable
from woodwork.demo import load_retail
from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage
)


@pytest.fixture(autouse=True)
def set_testing_headers():
    opener = urllib.request.build_opener()
    opener.addheaders = [('Testing', 'True')]
    urllib.request.install_opener(opener)


def test_load_retail_diff():
    nrows = 10
    df = load_retail(nrows=nrows, return_dataframe=True)
    assert df.shape[0] == nrows
    nrows_second = 11
    df = load_retail(nrows=nrows_second, return_dataframe=True)
    assert df.shape[0] == nrows_second

    assert 'order_product_id' in df.columns
    assert df['order_product_id'].is_unique


def test_load_retail_datatable():
    dt = load_retail(nrows=10, return_dataframe=False)
    assert isinstance(dt, DataTable)

    expected_logical_types = {
        'order_product_id': Categorical,
        'order_id': Categorical,
        'product_id': Categorical,
        'description': NaturalLanguage,
        'quantity': Integer,
        'order_date': Datetime,
        'unit_price': Double,
        'customer_name': Categorical,
        'country': Categorical,
        'total': Double,
        'cancelled': Boolean,
    }

    for column in dt.columns.values():
        assert column.logical_type == expected_logical_types[column.name]

    assert dt.index == 'order_product_id'
    assert dt.time_index == 'order_date'
    assert dt.columns['order_product_id'].semantic_tags == {'index'}
    assert dt.columns['order_date'].semantic_tags == {'time_index'}
